import os
import yaml
from functools import reduce
from typing import Dict, Any

default_config_file = 'mape.yml'
config_dict = {}


def set_config(config: Dict):
    global config_dict
    config_dict = config


def load(file: str = None):
    try:
        file = file or default_config_file
        with open(file, 'r') as config_file:
            config = yaml.load(config_file, Loader=yaml.Loader)
            set_config(config)
            return config
    except FileNotFoundError as e:
        return False


def get(path: str, default: Any = None, cfg: Dict = None):
    keys = path.split('.')
    cfg = cfg or config_dict

    ret = reduce(lambda c, k: c.get(k, {}), keys, cfg)
    return default if isinstance(ret, Dict) and not len(ret) else ret


def set(path: str, value, cfg = None):
    if value is None:
        return None

    keys = path.split('.')
    cfg = cfg or config_dict

    for key in keys[:-1]:
        cfg = cfg.setdefault(key, {})

    cfg[keys[-1]] = value
    return value


# Load from mape lib directory
default = load(os.path.dirname(os.path.realpath(__file__)) + '/mape.default.yml')
set_config(default)



import sys
import uuid
import inspect
import logging
import asyncio
import threading

import functools
from dataclasses import dataclass

_datefmt = '%H:%M:%S'
# _datefmt='%d %H:%M:%S'
_fmt = '%(asctime)s.%(msecs)-3d %(threadShortName)-3s %(name)-30s %(levelname)-8s %(message)s'
#_fmt = '%(relativeCreated)-6d %(threadShortName)-3s %(name)-20s: %(levelname)-8s %(message)s'


def cur_thread():
    """
    Return a unique number for the current thread'
    """
    n = threading.current_thread().name
    if 'Main' in n:
        return 'M'
    return '%s' % ('T' + n.rsplit('-', 1)[-1])


def log_record_enrich(record):
    record.threadShortName = cur_thread()
    return True


def init_logger(name: str = None, lvl: int = logging.WARNING, fmt: str = _fmt) -> logging.Logger:
    """
    Set base/root level logger for a package name
    :param name: base logger name (ie. package name)
    :param lvl: default log level
    :param fmt: format log string
    :return: logger
    """
    formatter = logging.Formatter(fmt, datefmt=_datefmt)

    # Handler for stdout, filtered for log lvl lower than INFO
    h_out = logging.StreamHandler(sys.stdout)
    h_out.setLevel(logging.DEBUG)
    h_out.addFilter(lambda record: record.levelno <= logging.INFO)
    h_out.setFormatter(formatter)
    h_out.addFilter(log_record_enrich)

    # Handler for stderr, filtered for log lvl greater than WARNING
    h_err = logging.StreamHandler(sys.stderr)
    h_err.setLevel(logging.WARNING)
    h_err.setFormatter(formatter)
    h_err.addFilter(log_record_enrich)

    # Get (and mod) the base/root logger (eg. package name)
    logger = logging.getLogger(name)
    logger.setLevel(lvl)

    logger.addHandler(h_out)
    logger.addHandler(h_err)

    return logger


from rx.core import Observer


def caller_module_name(depth=1):
    # since we get this function caller
    depth += 1
    frame_info = inspect.stack()[depth]
    mod = inspect.getmodule(frame_info.frame)
    return mod.__name__ if mod else ''


class LogObserver(Observer):
    def __init__(self, postfix=None, module_name=None):
        self.prefix = f"({str(hash(self))[-4:]})"
        self.postfix = f" | {postfix}" if postfix else ''

        if module_name is None:
            # f.__globals__.get("_LOGGER", _LOGGER)
            self.logger = logging.getLogger(caller_module_name())
        else:
            self.logger = logging.getLogger(module_name)

        if not self.logger.hasHandlers():
            self.logger.info = print
            self.logger.error = print

    def on_next(self, value):
        self.logger.info(f"{self.prefix} on next: {value}{self.postfix}")

    def on_error(self, error):
        self.logger.error(f"on error: {error}{self.postfix}")

    def on_completed(self):
        self.logger.info(f"completed{self.postfix}")


@dataclass
class GenericObject:
    pass


def generate_uid(collision_set=None, prefix=None):
    while True:
        uid = f"{prefix}{str(uuid.uuid4())[:8]}"
        if not collision_set or uid not in collision_set:
            break

    return uid


async def task_exception(awaitable, module_name=None):
    """
    Wrap coro() passed to asyncio.crete_task(), allow to catch exception during task execution
    note: https://bugs.python.org/issue39839

    Examples:
        >>> task = asyncio.create_task(task_exceptions(coro(*args, *kwargs))
        ...

    :param awaitable: coro() (coroutine object: an object returned by calling a coroutine function)
    :param module_name: name used for get logger
    :return: wrapped coro()
    """
    module_name = module_name or caller_module_name(5)
    logger = logging.getLogger(module_name)

    try:
        return await awaitable
    except Exception as e:
        logger.exception(e)


def log_task_exception(coro):
    """
    Decorator of task_exception(). Use on coro function instead of
    :param coro: coroutine function
    :return: wrapped coro()
    """
    module_name = caller_module_name()

    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        return await task_exception(coro(*args, **kwargs), module_name=module_name)
    return wrapped


def auto_task(func, *args, **kwargs):
    if asyncio.iscoroutinefunction(func):
        awaitable = task_exception(func(*args, **kwargs))
        return asyncio.create_task(awaitable)
    else:
        return func(*args, **kwargs)


async def aio_call(func):
    """
    Call and return value independently func is a coro or classic func.
    note: of course must be used inside an async func.

    Examples:
        >>> await auto_call(func(args, kwargs))
        >>> await auto_call(coro(args, kwargs))
        ...
    """
    if inspect.iscoroutine(func):
        return await func
    else:
        return func


def setdefaultattr(obj, name, value):
    """ As Dict.setdefault but for object """
    try:
        return getattr(obj, name)
    except AttributeError:
        setattr(obj, name, value)
    return value



from __future__ import annotations

import json
import pickle
from typing import Type, Any, List

from purse.collections import T
from pydantic import BaseModel


class Pickled:
    pass


def obj_from_raw(value_type: Type[T], raw_item: str | bytes) -> T | Any:
    if issubclass(value_type, BaseModel):
        return value_type.parse_raw(raw_item)
    elif issubclass(value_type, Pickled):
        return pickle.loads(raw_item)
    elif issubclass(value_type, dict):
        return json.loads(raw_item)
    elif issubclass(value_type, str):
        if isinstance(raw_item, bytes):
            return raw_item.decode()
        else:
            return raw_item
    else:
        if isinstance(raw_item, str):
            return raw_item.encode()
        else:
            return raw_item


def list_from_raw(value_type: Type[T], raw_list: List[Any]) -> List[T]:
    obj_list: List[T] = []
    if issubclass(value_type, BaseModel):
        for raw_item in raw_list:
            obj: Any = value_type.parse_raw(raw_item)
            obj_list.append(obj)
        return obj_list
    elif issubclass(value_type, Pickled):
        for raw_item in raw_list:
            obj: Any = pickle.loads(raw_item)
            obj_list.append(obj)
    elif issubclass(value_type, dict):
        for raw_item in raw_list:
            obj = json.loads(raw_item)
            obj_list.append(obj)
        return obj_list
    elif issubclass(value_type, str):
        for raw_item in raw_list:
            if isinstance(raw_item, bytes):
                obj = raw_item.decode()
                obj_list.append(obj)
            else:
                obj_list.append(raw_item)
        return obj_list
    else:
        for raw_item in raw_list:
            if isinstance(raw_item, str):
                obj = raw_item.encode()
                obj_list.append(obj)
            else:
                obj_list.append(raw_item)
        return obj_list


def obj_to_raw(value_type: Type[T], value: T) -> str | bytes:
    if isinstance(value, BaseModel) and isinstance(value, value_type):
        assert isinstance(value, BaseModel)
        return value.json()
    elif value_type is Pickled:
        return pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
    elif isinstance(value, dict):
        return json.dumps(value)
    elif isinstance(value, (str, bytes)):
        return value
    else:
        raise ValueError(
            f"Incorrect type. Expected type: {value_type} "
            f"while give Value type {type(value)}")



from enum import Enum
from typing import Callable
from functools import partial

from fastapi import FastAPI, HTTPException, Depends
from starlette.requests import Request
from starlette.responses import Response

import mape
from mape.loop import Loop
from mape.base_elements import Element

from ..de_serializer import obj_from_raw, Pickled


class Port(str, Enum):
    p_in = 'in'
    p_out = 'out'


class Notification(str, Enum):
    next = 'next'
    error = 'error'
    completed = 'completed'


element_notify_path = '/loops/{loop_uid}/elements/{element_uid}'


def api_setup(fastapi_app: FastAPI, mape_app: mape.App, deserializer: Callable = None):
    deserializer = deserializer or partial(obj_from_raw, Pickled)

    def common_loop(loop_uid: str) -> Loop:
        if loop_uid not in mape_app.loops:
            raise HTTPException(status_code=400, detail=f"Loop '{loop_uid}' does not exist")

        return mape_app.loops[loop_uid]

    def common_element(element_uid: str, loop: Loop = Depends(common_loop)) -> Element:
        if element_uid not in mape_app.loops[loop.uid]:
            raise HTTPException(status_code=400, detail=f"Element '{element_uid}' does not exist")

        return mape_app.loops[loop.uid][element_uid]

    @fastapi_app.get('/loops', tags=['loops'],
                     summary='Retrive Loops list', response_description='The created item')
    async def get_loops():
        return list(mape_app.loops.keys())

    @fastapi_app.get('/loops/{loop_uid}/elements',
                     tags=['elements'], summary='Retrive Elements list by Loop', response_description='The created item')
    async def get_elements(loop: common_loop = Depends()):
        return list(loop.elements.keys())

    @fastapi_app.get('/levels',
                     tags=['levels'], summary='Retrive Levels list', response_description='The created item')
    async def get_levels():
        return list(mape_app.levels.keys())

    @fastapi_app.post(element_notify_path,
                      tags=['elements'],
                      summary='Push stream on Element Port',
                      response_description='The created item')
    async def element_notify(request: Request,
                             element: common_element = Depends(),
                             port: Port = Port.p_in,
                             notification: Notification = Notification.next):
        """
        Create an item with all the information:

        - **name**: each item must have a name
        - **description**: a long description
        - **price**: required
        - **tax**: if the item doesn't have tax, you can omit this
        - **tags**: a set of unique tag strings for this item
        """
        port = element.port_in if port is Port.p_in else element.port_out

        body = deserializer(await request.body())

        if notification is Notification.next:
            port.on_next(body)
        elif notification is Notification.error:
            port.on_error(body)
        elif notification is Notification.completed:
            port.on_completed()



from __future__ import annotations

import asyncio
import aioredis
import logging
from functools import partial
from typing import Tuple, List, Dict, Callable

import mape
from mape.remote.de_serializer import obj_from_raw, Pickled
from mape.utils import log_task_exception, auto_task

logger = logging.getLogger(__name__)

def subscribe_handler_deco(channels_patterns, full_message=False, deserializer=None, redis=None):
    channels_patterns = channels_patterns if isinstance(channels_patterns, List) else [channels_patterns]

    def decorator(func):
        sub_handlers = {pattern: func for pattern in channels_patterns}
        task = subscribe_handler(sub_handlers, full_message, deserializer, redis)

        func.task = task
        func.cancel = task.cancel
        return func

    return decorator


def subscribe_handler(sub_handlers: Dict[str, Callable], full_message=False, deserializer=None, redis=None):
    redis = redis or mape.redis

    if not isinstance(redis, aioredis.Redis):
        logger.error("You are trying to use Redis without config it!")

    deserializer = deserializer or partial(obj_from_raw, Pickled)

    def _on_publish(message, callback):
        message['data'] = deserializer(message['data'])
        callback(message if full_message else message['data'])

    def on_cancel(_):
        sub_channels = sub_handlers.keys()
        asyncio.create_task(pubsub.unsubscribe(sub_channels))

    @log_task_exception
    async def init_redis_sub():
        patterns_callbacks = {pattern: partial(_on_publish, callback=handler) for pattern, handler in
                              sub_handlers.items()}
        await pubsub.psubscribe(**patterns_callbacks)
        await pubsub.run()

    pubsub = redis.pubsub()
    task = asyncio.create_task(init_redis_sub())
    task.add_done_callback(on_cancel)
    return task


def notifications_handler(handler: Callable, key: str, cmd_filter=(), full_message=False, *args, **kwargs):
    cmd_filter = cmd_filter if isinstance(cmd_filter, (Tuple, List)) else [cmd_filter]

    def _pre_handler(message):
        redis_cmd = message['data'] if full_message else message

        if not cmd_filter or redis_cmd in cmd_filter:
            auto_task(handler, message)

    subscribe_handler({key: _pre_handler}, full_message, deserializer=partial(obj_from_raw, str), *args, **kwargs)



from __future__ import annotations

from aioredis import Redis
from typing import Any, Dict, Type, Union, Tuple, Iterable, List, TypeVar

from mape.loop import Loop
from mape.base_elements import Element
from mape.level import Level
from mape.knowledge import Knowledge
from mape.utils import generate_uid
from mape.constants import RESERVED_PREPEND, RESERVED_SEPARATOR


class App:
    uid: str = 'app'

    def __init__(self, redis: Redis) -> None:
        self._redis: Redis = redis
        self._loops: Dict[str, Loop] = dict()
        self._levels: Dict[str, Level] = dict()
        self._k = Knowledge(self._redis, f"k{RESERVED_SEPARATOR}{self.uid}")

    def add_loop(self, loop):
        uid = loop.uid or generate_uid(self._loops, prefix=loop.prefix)

        if self.has_loop(uid) or hasattr(self, uid):
            return False

        loop._uid = uid
        loop._app = self
        self._loops[uid] = loop

        return uid

    register = add_loop

    def has_loop(self, loop) -> bool:
        uid = loop.uid if hasattr(loop, 'uid') else loop
        return uid in self._loops

    def __contains__(self, loop):
        return self.has_loop(loop)

    def __getattr__(self, uid):
        """ Allow access (through dot notation) to mape loops.
        note: __getattr__() is called only when no real object attr exist. """
        if uid in self._loops:
            return self._loops[uid]

        super().__getattribute__(uid)

    def __getitem__(self, path: str) -> Loop | Element | Level:
        items = path.split(RESERVED_SEPARATOR)
        count_items = len(items)

        if count_items == 1:
            # path: 'loop_uid'
            try:
                return self._loops[items[0]]
            except KeyError as err:
                raise KeyError(f"Loop '{items[0]}' not exist")
        elif count_items == 2:
            # path: 'loop_uid.element_uid'
            loop = self[items[0]]
            return loop[items[1]]
        elif count_items == 3:
            # path: 'level_uid.loop_uid.element_uid'
            try:
                level = self._levels[items[0]]
            except KeyError as err:
                raise KeyError(f"Level '{items[0]}' not exist")
            return level[items[1]]

        raise KeyError(f"Path is malformed {path}")

    def __iter__(self):
        return iter(self._loops.values())

    def add_default_level(self, level_uid: str):
        """
        Create and add the new level only if not already exist
        ie. no conflict management like loop, if already exist return that
        """
        if level_uid not in self._levels:
            self._levels[level_uid] = Level(level_uid, app=self)

        return self._levels[level_uid]

    @property
    def redis(self):
        return self._redis

    @property
    def loops(self):
        return self._loops

    @property
    def levels(self):
        return self._levels

    @property
    def k(self) -> Knowledge:
        return self._k



from __future__ import annotations

import logging
import json
import aiohttp
import asyncio

from typing import Any
from functools import partial
from aiohttp.client_exceptions import ClientError
from rx.core import Observer, Observable

import mape
from mape.utils import log_task_exception, task_exception
from mape.constants import RESERVED_SEPARATOR

from .api import Port, Notification, element_notify_path
from ..de_serializer import obj_to_raw, obj_from_raw, Pickled

logger = logging.getLogger(__name__)

# TODO:
# * better implemented with a queue (like PubObserver) to preserve notifications order


def _element_path2url_path(element_path):
    loop_uid, element_uid = element_path.split(RESERVED_SEPARATOR)
    return element_notify_path.format(loop_uid=loop_uid, element_uid=element_uid)


class POSTObserver(Observer):
    def __init__(self,
                 base_url: str,
                 path: str,
                 port: Port = Port.p_in,
                 serializer=None,
                 session: aiohttp.ClientSession = None) -> None:
        self._base_url = base_url

        try:
            self._path = path if path.startswith('/') else _element_path2url_path(path)
        except ValueError as e:
            logger.error(f"Malformed element_path: '{path}'")

        self._port = port
        self._session = session or aiohttp.ClientSession(base_url)
        self._serializer = serializer or partial(obj_to_raw, Pickled)
        # self._queue = asyncio.Queue()

        super().__init__()

    def _on_next_core(self, value: Any) -> None:
        asyncio.create_task(self.post(value, Notification.next))

    def _on_error_core(self, error: Exception) -> None:
        asyncio.create_task(self.post(error, Notification.error))

    def _on_completed_core(self) -> None:
        asyncio.create_task(self.post(None, Notification.completed))

    @log_task_exception
    async def post(self, value, notification: Notification):
        try:
            data = self._serializer(value)
            params = {'port': self._port.value, 'notification': notification.value}

            async with self._session.post(self._path, data=data, params=params) as resp:
                if resp.status != 200:
                    text = await resp.text()
                    logger.error(f"Response from '{self._path}' status {resp.status}: '{json.loads(text)}'")

        except aiohttp.client_exceptions.ClientError as e:
            logger.error(e)

    def dispose(self) -> None:
        asyncio.create_task(task_exception(self._session.close()))
        super().dispose()

    def __del__(self):
        self.dispose()


# TODO: to finish
import rx
from rx.disposable import Disposable
from fastapi import FastAPI
from starlette.requests import Request
from starlette.responses import Response


class PostObservable(Observable):
    def __init__(self, path, deserializer=None, fastapi: FastAPI | None = None) -> None:
        self._path = path
        self._fastapi = fastapi or mape.fastapi
        self._deserializer = deserializer or partial(obj_from_raw, Pickled)

        self._task = None

        def on_subscribe(observer, scheduler):

            async def on_next1(request: Request):
                print(await request.body())
                return await request.body()
                # observer.on_next(value)

            self._fastapi.post(self._path, response_class=Response)(on_next1)

            print("subscribed", self._path)

            return Disposable()

        self._auto_connect = rx.create(on_subscribe).pipe(ops.dematerialize(), ops.share())
        super().__init__()

    def _subscribe_core(self, observer, scheduler=None):
        return self._auto_connect.subscribe(observer, scheduler=scheduler)

    # TODO: re-running create unpredictable error (endpoint appear and disappear
    # def close(self):
    #     # Find the route by path
    #     # try:
    #     route = [route for route in self._fastapi.routes if route.path == self._path]
    #     if len(route):
    #         self._fastapi.routes.remove(route[0])
    #     # except (IndexError, ValueError) as e:
    #     #     # already removed (ie. webserver shutdown)
    #     #     pass
    #
    # def __del__(self):
    #     self.close()


from aioredis import Redis
from typing import Any, Dict, Type, Union, Tuple, Iterable, List, TypeVar, Callable
from functools import partial

from mape.remote.redis import (
    RedisKeySpace,
    RedisHash,
    RedisSet,
    RedisList,
    RedisKey,
    RedisSortedSet,
    RedisPriorityQueue,
    RedisQueue,
    RedisLifoQueue,
    Redlock,
    notifications_handler
)
from mape.remote.de_serializer import Pickled, obj_from_raw
from mape.constants import RESERVED_PREPEND, RESERVED_SEPARATOR

T = TypeVar('T')


class Knowledge:
    def __init__(self, redis: Redis, prefix: str) -> None:
        self._redis: Redis = redis
        self._prefix: str = prefix + RESERVED_SEPARATOR

        self._keyspace = self.create_keyspace(RESERVED_PREPEND + 'default_keyspace', value_type=Pickled)

    def create_keyspace(self, key: str, value_type: Type[T]):
        return RedisKeySpace(self._redis, self._prefix + key + RESERVED_SEPARATOR, value_type=value_type)

    def create_hash(self, key: str, value_type: Type[T]) -> RedisHash[T]:
        return RedisHash(self._redis, self._prefix + key, value_type=value_type)

    def create_set(self, key: str, value_type: Type[T]) -> RedisSet[T]:
        return RedisSet(self._redis, self._prefix + key, value_type=value_type)

    def create_list(self, key: str, value_type: Type[T]) -> RedisList[T]:
        return RedisList(self._redis, self._prefix + key, value_type=value_type)

    def create_sortedset(self, key: str, value_type: Type[T]) -> RedisSortedSet[T]:
        return RedisSortedSet(self._redis, self._prefix + key, value_type=value_type)

    def create_priorityqueue(self, key: str, value_type: Type[T]) -> RedisPriorityQueue[T]:
        return RedisPriorityQueue(self._redis, self._prefix + key, value_type=value_type)

    def create_queue(self, key: str, value_type: Type[T]) -> RedisQueue[T]:
        return RedisQueue(self._redis, self._prefix + key, value_type=value_type)

    def create_lifoqueue(self, key: str, value_type: Type[T]) -> RedisLifoQueue[T]:
        return RedisLifoQueue(self._redis, self._prefix + key, value_type=value_type)

    def create_lock(self, key, masters: List[Redis], *args, **kwargs):
        return Redlock(key, masters, *args, **kwargs)

    def notifications(self, handler: Callable, key: str, *args, **kwargs):
        return notifications_handler(handler, f"__keyspace@*__:{self._prefix}{key}", *args, **kwargs)

    @property
    def keyspace(self) -> RedisKeySpace[Pickled]:
        return self._keyspace

    @property
    def prefix(self):
        return self._prefix



from __future__ import annotations

from typing import Any, Callable, Optional, Union, Awaitable, Coroutine, NamedTuple, Tuple

import rx
from rx.subject import Subject
from rx.core import Observer, Observable
from rx.disposable import Disposable, CompositeDisposable
from rx.operators import *

import mape
from .base_elements import Element
from .typing import Mapper, OpsChain, DestMapper, Message

MapeElement = Union[Subject, Element]


def through(subject: MapeElement):
    """ Item pass through the subject """

    def _through(source):
        def subscribe(observer, scheduler=None):
            source_dispose = source.subscribe(subject, scheduler=scheduler)
            subject_dispose = subject.subscribe(observer, scheduler=scheduler)

            return CompositeDisposable(source_dispose, subject_dispose)

        return Observable(subscribe)

    return _through


sitm = through


def group_and_pipe(operators: OpsChain, key_mapper: Mapper = None):
    key_mapper = key_mapper or (lambda item: item.src)
    ops_in = list(pipe) if isinstance(pipe, Tuple) else [pipe]

    def _group_and_pipe(source):
        def subscribe(observer, scheduler=None):
            return source.pipe(
                group_by(key_mapper),
                flat_map(
                    lambda grp: grp.pipe(*operators)
                )).subscribe(
                observer.on_next,
                observer.on_error,
                observer.on_completed,
                scheduler)

        return rx.create(subscribe)

    return _group_and_pipe


reverse_proxy = group_and_pipe


def _gateway_dest_mapper(item: Message):
    return mape.app[item.dst]


def gateway(dest_mapper: DestMapper = None):
    dest_mapper = dest_mapper or _gateway_dest_mapper

    def _gateway(source):
        def subscribe(observer, scheduler=None):
            from collections.abc import Iterable

            def on_next(item):
                dests = dest_mapper(item)
                if not isinstance(dests, Iterable):
                    # Convert dest in a tuple
                    dests = (dests,)

                for dest in dests:
                    dest.on_next(item)
                observer.on_next(item)

            def on_error(error):
                observer.on_error(error)

            def on_completed():
                observer.on_completed()

            return source.subscribe(
                on_next,
                on_error,
                on_completed,
                scheduler)

        return rx.create(subscribe)

    return _gateway


router = gateway




from __future__ import annotations

import logging
from typing import Callable, Dict, Any, Mapping, Tuple, List, Iterable

from rx.core import Observer
from influxdb_client import InfluxDBClient, Point, WriteOptions
from influxdb_client.client.write_api import SYNCHRONOUS, ASYNCHRONOUS

from mape.typing import Message

logger = logging.getLogger(__name__)


def _fields_mapper(item, key='value'):
    if hasattr(item, key):
        value = getattr(item, key)
    elif isinstance(item, Mapping):
        value = item.get(key)
    else:
        value = item

    return key, value


class InfluxObserver(Observer):
    def __init__(self,
                 measurement: str | None = None,
                 tags: Iterable | Iterable[Iterable] | None = None,
                 fields_mapper: Callable | None = None,
                 bucket: str | None = None,
                 is_raw: bool = False,
                 write_options: WriteOptions = ASYNCHRONOUS,
                 client: InfluxDBClient | None = None
                 ) -> None:
        from . import _config

        self._measurement = measurement
        self._tags = tags
        self._fields_mapper = fields_mapper or _fields_mapper
        self._bucket = bucket or _config['bucket']
        self._is_raw = is_raw
        self._write_options = write_options
        self._org = _config['org']
        self._client = client or InfluxDBClient(
            **{k: _config[k] for k in ('url', 'token', 'org', 'debug') if k in _config}
        )
        self._write_api = self._client.write_api(write_options=write_options)

        if self._tags and not isinstance(self._tags[0], (Tuple, List)):
            self._tags = (self._tags,)

        super().__init__()

    def _on_next_core(self, item: Any) -> None:
        if self._is_raw:
            self._write_api.write(self._bucket, self._org, record=item)
        else:
            measurement = self._measurement or type(item).__name__
            point = Point(measurement)

            if not self._tags and hasattr(item, '__dict__'):
                value_as_dict = item.__dict__
                tags = [(k, value_as_dict[k]) for k in value_as_dict if k not in ('value', 'timestamp')]
            else:
                tags = self._tags or list()

            for tag, value in tags:
                point.tag(tag, value)

            fields = self._fields_mapper(item)
            if not isinstance(fields[0], (Tuple, List)):
                fields = (fields,)

            for field, value in fields:
                point.field(field, value)

            logger.debug(f"InfluxDB write: {point.to_line_protocol()}")
            self._write_api.write(self._bucket, self._org, record=point)

    def dispose(self) -> None:
        self._client.__del__()
        super().dispose()

    def __del__(self):
        self.dispose()


import mape
from mape.constants import RESERVED_PREPEND, RESERVED_SEPARATOR
from mape.knowledge import Knowledge


class Level:
    """
    Level is created only by application usually after a Loop request.
    It aggregates Loops, and is used for Knowledge namespace.
    Loop uid still be unique, and can be directly accessed by App.
    If a Level (ie uid) already exist not throw exception/conflict (like Loop)
    but simply return the existed one (see App.add_default_level())
    """

    def __init__(self, uid: str, app=None) -> None:
        self._uid = uid
        self._app = app or mape.app
        self._loops = dict()

        self._k = Knowledge(self.app.redis, f"k{RESERVED_SEPARATOR}level{RESERVED_SEPARATOR}{self.uid}")

    def add_loop(self, loop):
        uid = loop.uid

        if self.has_loop(uid) or hasattr(self, uid):
            return False

        if loop in loop.app.levels:
            del loop.app.levels.loops[uid]

        loop._level = self
        self._loops[uid] = loop

        return uid

    register = add_loop

    def has_loop(self, loop) -> bool:
        uid = loop.uid if hasattr(loop, 'uid') else loop
        return uid in self._loops

    def __contains__(self, loop):
        return self.has_loop(loop)

    def __getattr__(self, uid):
        """ Allow access (through dot notation) to mape loops.
        note: __getattr__() is called only when no real object attr exist. """
        if uid in self._loops:
            return self._loops[uid]

        super().__getattribute__(uid)

    def __getitem__(self, path: str):
        items = path.split(RESERVED_SEPARATOR)
        count_items = len(items)

        if count_items == 1:
            # path: 'loop_uid'
            try:
                return self._loops[items[0]]
            except KeyError as err:
                raise KeyError(f"Loop '{items[0]}' not exist in level '{self.uid}'")
        elif count_items == 2:
            # path: 'loop_uid.element_uid'
            loop = self[items[0]]
            return loop[items[1]]

        raise KeyError(f"Path is malformed {path}")

    def __iter__(self):
        return iter(self._loops.values())

    @property
    def uid(self):
        return self._uid

    def __str__(self) -> str:
        return self.uid

    @property
    def app(self):
        return self._app

    @property
    def loops(self):
        return self._loops

    @property
    def k(self):
        return self._k



from __future__ import annotations

from fastapi import FastAPI
from textwrap import dedent

import mape
from .api import api_setup

fastapi: FastAPI | None = None

description = dedent('''
        MAPE framework RESTfull API interface
       
        ## Levels
        
        You can **read items**.
        
        ## Loops
        
        You can **read items**.
        
        ## Elements
        
        You will be able to:
        
        * **Levels** (_not implemented_).
        * **Post** (_not implemented_).''')

tags_metadata = [
    {
        'name': 'levels',
        'description': 'Operation with levels',
    },
    {
        'name': 'loops',
        'description': 'Operations with loops',
    },
    {
        'name': 'elements',
        'description': 'Operations with Elements',
    },
]


def setup(mape_app: mape.App, version):
    global fastapi

    if not fastapi:
        fastapi = FastAPI(
            title='MAPE framework',
            description=description,
            version=version,
            terms_of_service='http://example.com/terms/',
            contact={
                'name': 'Emanuele Palombo',
                'url': 'https://github.com/elbowz',
                # 'email': 'ask@x-force.example.com',
            },
            license_info={
                'name': 'GPL-3.0 License',
                'url': 'https://www.gnu.org/licenses/gpl-3.0.en.html',
            },
            openapi_tags=tags_metadata
        )

        api_setup(fastapi, mape_app)

    return fastapi




from purse import (
    RedisKeySpace,
    RedisHash,
    RedisSet,
    RedisList,
    RedisKey,
    RedisSortedSet,
    RedisPriorityQueue,
    RedisQueue,
    RedisLifoQueue,
    Redlock
)

from .rx_utils import PubObserver, SubObservable
from .pubsub import subscribe_handler_deco, subscribe_handler, notifications_handler
from .collections_patch import purse_monkey_patch as _purse_monkey_patch

_purse_monkey_patch()



from __future__ import annotations

import functools
import logging
import inspect
from functools import wraps, partial
from typing import Type, Any, Tuple, List, Dict, Callable, Optional, Union, Awaitable, Coroutine, NamedTuple, TypeVar
import types

import mape

from mape.base_elements import Element, Monitor, Analyze, Plan, Execute, UID, to_element_cls, make_func_class
from mape.knowledge import Knowledge
from mape.utils import generate_uid
from mape.typing import MapeLoop, OpsChain
from mape.constants import RESERVED_PREPEND, RESERVED_SEPARATOR

logger = logging.getLogger(__name__)

TElement = TypeVar('TElement')


class Loop(MapeLoop):
    prefix: str = 'l_'

    def __init__(self, uid: str = None, level: str = '', app=None) -> None:
        self._uid = uid
        self._elements: Dict[str, Element] = dict()
        self._app: mape.application.App = app or mape.app

        # Add Loop to App
        self._uid = self.add_to_app(self._app)

        if not self._uid:
            raise ValueError(f"'{uid}' already taken, or name is protected")

        self._level = self._app.add_default_level(level)

        if not self.add_to_level(self._level):
            raise ValueError(f"'{uid}' name is protected")

        self._k: Knowledge = Knowledge(self.app.redis, f"k{RESERVED_SEPARATOR}loop{RESERVED_SEPARATOR}{self.uid}")

    def add_to_app(self, app):
        return app.add_loop(self)

    move_to_app = add_to_app

    def add_to_level(self, level):
        return level.add_loop(self)

    move_to_level = add_to_app

    def add_element(self, element):
        uid = element.uid or generate_uid(self._elements, prefix=element.prefix)

        if self.has_element(uid) or hasattr(self, uid):
            return False

        if element in element._loop:
            del element.loop.elements[element.uid]

        element._uid = uid
        element._loop = self
        self._elements[uid] = element

        return uid

    def has_element(self, element):
        uid = element.uid if hasattr(element, 'uid') else element
        return uid in self._elements

    def __contains__(self, element):
        return self.has_element(element)

    def __getattr__(self, uid):
        """ Allow access (through dot notation) to mape elements.
        note: __getattr__() is called only when no real object attr exist. """
        if uid in self._elements:
            return self._elements[uid]

        super().__getattribute__(uid)

    def __getitem__(self, uid) -> Element:
        try:
            return self._elements[uid]
        except KeyError as err:
            raise KeyError(f"Element '{uid}' not exist in '{self._uid}' loop")

    def __iter__(self):
        return iter(self._elements.values())

    @property
    def uid(self):
        return self._uid

    def __str__(self) -> str:
        return self.uid

    @property
    def app(self):
        return self._app

    @property
    def level(self):
        return self._level

    @property
    def k(self) -> Knowledge:
        return self._k

    @property
    def elements(self):
        return self._elements

    def monitor(self,
                func: Callable = None, /, *,
                uid: str | UID = UID.DEF,
                ops_in: Optional[OpsChain] = (),
                ops_out: Optional[OpsChain] = ()
                ) -> Monitor:
        """ Function decorator """

        return self.add_func(func,
                             element_class=Monitor,
                             uid=uid,
                             ops_in=ops_in,
                             ops_out=ops_out)

    def analyze(self,
                func: Callable = None, /, *,
                uid: str | UID = UID.DEF,
                ops_in: Optional[OpsChain] = (),
                ops_out: Optional[OpsChain] = ()
                ) -> Analyze:
        """ Function decorator """

        return self.add_func(func,
                             element_class=Analyze,
                             uid=uid,
                             ops_in=ops_in,
                             ops_out=ops_out)

    def plan(self,
             func: Callable = None, /, *,
             uid: str | UID = UID.DEF,
             ops_in: Optional[OpsChain] = (),
             ops_out: Optional[OpsChain] = ()
             ) -> Plan:
        """ Function decorator """

        return self.add_func(func,
                             element_class=Plan,
                             uid=uid,
                             ops_in=ops_in,
                             ops_out=ops_out)

    def execute(self,
                func: Callable = None, /, *,
                uid: str | UID = UID.DEF,
                ops_in: Optional[OpsChain] = (),
                ops_out: Optional[OpsChain] = ()
                ) -> Execute:
        """ Function decorator """

        return self.add_func(func,
                             element_class=Execute,
                             uid=uid,
                             ops_in=ops_in,
                             ops_out=ops_out)

    # Alternative method to declare execute, but missing signature/type hint
    # execute_test = functools.partialmethod(add_func, element_class=Execute)

    # TODO: provare con l'overload una per ElementFunc e una per tutte le altre classi
    def register(self,
                 cls: Type[TElement] = None,
                 /, *,
                 uid: str | UID = UID.DEF,
                 ops_in: Optional[OpsChain] = (),
                 ops_out: Optional[OpsChain] = ()
                 ) -> Type[TElement] | TElement:
        """ Class decorator """
        if cls is None:
            return partial(self.register,
                           uid=uid,
                           ops_in=ops_in,
                           ops_out=ops_out)

        kwargs = {}

        args_name = inspect.signature(cls.__init__).parameters.keys()

        # Append (if present in the signature) additional cls constructor args for obj initialization
        if 'uid' in args_name and cls.__name__ != 'ElementFunc':
            # Avoid if Element cls generated by func.
            # uid with UID.DEF already set with func name
            kwargs['uid'] = uid
        if 'ops_in' in args_name and ops_in:
            kwargs['ops_in'] = ops_in
        if 'ops_out' in args_name and ops_out:
            kwargs['ops_out'] = ops_out

        element = cls(loop=self, **kwargs)

        return element if cls.__name__ == 'ElementFunc' else cls

    register_cls = register

    def add_func(self,
                 func: Callable = None,
                 /, *,
                 element_class: Type[TElement] = Element,
                 uid: str | UID = UID.DEF,
                 ops_in: Optional[OpsChain] = (),
                 ops_out: Optional[OpsChain] = ()
                 ) -> TElement:
        """ Function decorator """

        if func is None:
            return partial(self.add_func,
                           element_class=element_class,
                           uid=uid,
                           ops_in=ops_in,
                           ops_out=ops_out)

        cls = make_func_class(func,
                              element_class=element_class,
                              default_uid=uid,
                              default_ops_in=ops_in,
                              default_ops_out=ops_out)

        return self.register(cls)



from __future__ import annotations

import signal
import aioredis
import warnings
from fastapi import FastAPI
from typing import Dict

from rx.scheduler.eventloop import AsyncIOScheduler

from . import config as mape_config
from .application import App
from .base_elements import *
from .loop import Loop
# from .operators import *
from .remote.rest import UvicornDaemon, setup as rest_setup
from .remote.influxdb import set_config as set_influxdb
from .utils import init_logger, task_exception

# Please make sure the version here remains the same as in pyproject.toml
__version__ = "0.1.0a3"

# Disable logging until user use logging.basicConfig()
# TODO: remove logger initialization
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

aio_loop: Optional[asyncio.AbstractEventLoop] = None
rx_scheduler: Optional[AsyncIOScheduler] = None
redis: Optional[aioredis.Redis] = None
uvicorn_webserver: Optional[UvicornDaemon] = None
fastapi: Optional[FastAPI] = None
app: Optional[App] = None
config: Optional[Dict] = None


def setup_logger():
    """
    Helper for quickly adding a StreamHandler to the logger.
    Useful for debugging and lib debugging
    """
    # This method needs to be in this __init__.py to get the __name__ correct
    init_logger(__name__)

    # Stop propagation on upper logger_name (eg. __main__ and logging.basicConfig())
    logging.getLogger(__name__).propagate = False


def stop():
    aio_loop.stop()
    uvicorn_webserver and uvicorn_webserver.stop()

    # Let's also cancel all running tasks:
    pending = asyncio.Task.all_tasks()

    for task in pending:
        logger.debug(f"Closing still running task: {task!r}")
        task.cancel()


def _start_web_server(host_port: str, loop):
    global uvicorn_webserver

    init_len_routes: Final = 4

    # Start Uvicorn webserver (forking a process) with the created routes by FastAPI
    if fastapi and (len_routes := len(fastapi.routes)) > init_len_routes:
        # The first 4 routes are instanced by default (ie swagger)
        from mape.remote.rest import UvicornDaemon

        host, port = host_port.split(':')
        uvicorn_webserver = UvicornDaemon(app=fastapi, loop=loop, host=host, port=port)

        system_path = fastapi.routes[:init_len_routes]
        mape_path = fastapi.routes[init_len_routes:]
        host_base_url = f"http://{uvicorn_webserver.config.host}:{uvicorn_webserver.config.port}"

        logger.info("APIs documentation:")
        for route in system_path:
            logger.info(f" * {host_base_url}{route.path}")

        logger.debug(f"Defined {len_routes - init_len_routes} REST MAPE endpoints:")
        for route in mape_path:
            logger.debug(f" * {host_base_url}{route.path}")

        logger.info('Webserver started: No more endpoint can be added since now!')


def init(debug=False, asyncio_loop=None, redis_url=None, rest_host_port=None, config_file=None):
    global config, aio_loop, rx_scheduler, redis, fastapi, app

    logging.getLogger(__name__).setLevel(logging.DEBUG if debug else logging.WARNING)

    config_file = config_file or mape_config.default_config_file

    if config := mape_config.load(config_file):
        logger.info(f"{config_file} loaded")
    else:
        logger.warning(f"{config_file} not found, using default config")
        config = mape_config.default

    mape_config.set('debug', debug)
    mape_config.set('redis.url', redis_url)
    mape_config.set('rest.host_port', rest_host_port)

    logger.debug(f"Config: {config}")

    # loop = asyncio.new_event_loop()
    # asyncio.set_event_loop(loop)
    # loop = asyncio_loop or loop or asyncio.get_event_loop()

    # TODO: check comment
    # We don't create a new event loop by default, because we want to be
    # sure that when this is called multiple times, each call of `init()`
    # goes through the same event loop. This way, users can schedule
    # background-tasks that keep running across multiple prompts.
    try:
        aio_loop = asyncio_loop or asyncio.get_event_loop()
    except RuntimeError:
        # Possibly we are not running in the main thread, where no event
        # loop is set by default. Or somebody called `asyncio.run()`
        # before, which closes the existing event loop. We can create a new
        # loop.
        aio_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(aio_loop)

    rx_scheduler = rx_scheduler or AsyncIOScheduler(aio_loop)

    if mape_config.get('redis.url'):
        redis = aioredis.from_url(redis_url, db=0)

    app = App(redis)

    if mape_config.get('rest.host_port'):
        fastapi = rest_setup(app, __version__)
        _start_web_server(rest_host_port, aio_loop)

    set_influxdb(mape_config.get('influxdb'))
    set_debug(mape_config.get('debug'))


def run(entrypoint: Union[Callable, Coroutine] = None):
    def on_done(_):
        # Catch stop execution (ie. ctrl+c or brutal stop)
        # notes: set the handler here to overwrite the Uvicorn handlers
        for signal_name in (signal.SIGINT, signal.SIGTERM):
            aio_loop.add_signal_handler(signal_name, stop)

        # loop.set_exception_handler(lambda *args: print("exception_handler", *args))

    if entrypoint is not None:
        if asyncio.iscoroutine(entrypoint):
            task = aio_loop.create_task(task_exception(entrypoint), name='entrypoint')
            task.add_done_callback(on_done)

    aio_loop.run_forever()


def set_debug(debug=False, asyncio_slow_callback_duration=0.1):
    # TODO: better debug and logging level support (it don't works as expected...)
    log_lvl = logging.DEBUG if debug else logging.WARNING

    # Enabling it, sometimes asyncio throw errors like:
    # "<_SelectorSocketTransport fd=11 read=polling write=<idle, bufsize=0>>: Fatal read error on socket transport"
    # aio_loop.set_debug(debug)

    logging.getLogger(__name__).setLevel(log_lvl)
    logging.getLogger('asyncio').setLevel(log_lvl)

    if fastapi:
        fastapi.debug = debug
        # Avoid double logging for uvicorn
        logging.getLogger('uvicorn').handlers.clear()

    # Threshold for "slow" tasks.
    # Reduce to smaller value to throw the error and understand the behaviour.
    # The default is 0.1 (100 milliseconds).
    aio_loop.slow_callback_duration = asyncio_slow_callback_duration

    if debug:
        # Report all mistakes managing asynchronous resources.
        warnings.simplefilter('always', ResourceWarning)
    else:
        warnings.simplefilter('ignore', ResourceWarning)


from __future__ import annotations

import logging
import inspect
import asyncio
from typing import Type, Any, List, Tuple, Callable, Optional, Union, Awaitable, Coroutine, NamedTuple, Final, overload, TypeVar
from enum import Flag, Enum
from functools import partial, wraps

import rx
from rx.subject import Subject
from rx.core import Observer, Observable, ConnectableObservable, typing as rx_typing
from rx.disposable import Disposable, CompositeDisposable
from rx import operators as ops
from dataclasses import dataclass, field

import mape
from mape import typing
from mape.constants import RESERVED_SEPARATOR
from .utils import init_logger, LogObserver, GenericObject, caller_module_name, task_exception, aio_call

logger = logging.getLogger(__name__)


class UID(Enum):
    RANDOM = None
    DEF = 1

    # TODO: remove, because never used
    @dataclass
    class MANUAL:
        value: str = ''


# TODO: use Pydantic (in general for dataclass and validation (ie also setting)
@dataclass
class Port:
    # TODO: add method to manipulate operators (append, prepend, etc...) and update pipe/input/subscribe?!
    input: Subject = None
    pipe: Observable = None
    operators: List[Callable[[Any], Any]] = field(default_factory=lambda: [])
    output: Subject = None
    disposable: Disposable = None


# TODO:
#  * inherit from Subject?
#  * single pipe from p_in to p_out and in the middle ops.through (ie on_next, on_error...)
#  * use a ConnectableObservable instead and exploit the .connect(), implementing .disconnect() ?!
#    inspiration: https://github.com/ReactiveX/RxPY/blob/release/v1.6.x/rx/backpressure/pausable.py
class Element(Observable, Observer, rx_typing.Subject):
    # TODO: pass as argument to _process_msg, instead of only move_on
    # evaluate if is it necessary:
    #  * is raise exception inside _process_msg the same that pass error
    #  * return instead of on_complete()? ...we need on_complete?!
    # @dataclass(frozen=True)
    # class ProcessFuncData:
    #     next: Callable
    #     error: Callable
    #     completed: Callable
    prefix: str = ''

    class Debug(Flag):
        DISABLE = 1
        IN = 2
        OUT = 4

    def __init__(self,
                 loop: typing.MapeLoop,
                 uid: str | UID = None,
                 ops_in: Optional[typing.OpsChain] = (),
                 ops_out: Optional[typing.OpsChain] = ()
                 ) -> None:
        uid = uid if uid != UID.DEF else self.__class__.__name__
        self._uid = uid if not hasattr(uid, 'value') else uid.value

        self._loop: mape.Loop = loop
        self._aio_loop = mape.aio_loop or asyncio.get_event_loop()
        self.is_running = False
        self._debug = GenericObject()

        # Add Element to loop
        self._uid = self.add_to_loop(loop)

        if not self._uid:
            raise ValueError(f"'{uid}' already taken, or name is protected")

        # Accept pipe(), ops(), (ops(),)
        # TODO: I think is not so good, and not always it works
        # hint:
        #   >>> seq = seq if isinstance(seq, (List, Tuple)) else (seq,)
        #   >>> sqe = seq if isinstance(seq, Sequence) else (seq,) # take also string (not good)
        #   >>> routes: typing.Sequence[BaseRoute] = None,
        #   >>> self.routes = [] if routes is None else list(routes)
        ops_in = list(ops_in) if isinstance(ops_in, Tuple) else [ops_in]
        ops_out = list(ops_out) if isinstance(ops_out, Tuple) else [ops_out]

        # Port in and out
        self._p_in = Port(input=Subject(), operators=ops_in)
        self._p_out = Port(input=Subject(), operators=ops_out, output=Subject())

        Observable.__init__(self)
        Observer.__init__(self, self._p_in.input.on_next, self._p_in.input.on_error, self._p_in.input.on_completed)

    async def _aio_init(self):
        """ Stuff to init with await """
        pass

    def __await__(self):
        return self._aio_init().__await__()

    def add_to_loop(self, loop: mape.Loop):
        return loop.add_element(self)

    move_to_loop = add_to_loop

    def debug(self, lvl: Element.Debug = Debug.DISABLE, module_name=None):
        module_name = module_name or caller_module_name()

        if Element.Debug.IN in lvl:
            self._debug.in_dispose = self._p_in.input.subscribe(
                LogObserver(f"in > [{self._loop.uid}{RESERVED_SEPARATOR}{self.uid}]", module_name)
            )
        elif hasattr(self._debug, 'in_dispose'):
            self._debug.in_dispose.dispose()

        if Element.Debug.OUT in lvl:
            self._debug.out_dispose = self._p_out.output.subscribe(
                LogObserver(f"[{self._loop.uid}{RESERVED_SEPARATOR}{self.uid}] > out", module_name)
            )
        elif hasattr(self._debug, 'out_dispose'):
            self._debug.out_dispose.dispose()

    def subscribe(self, observer: Optional[Union[rx_typing.Observer, rx_typing.OnNext]] = None,
                  on_error: Optional[rx_typing.OnError] = None, on_completed: Optional[rx_typing.OnCompleted] = None,
                  on_next: Optional[rx_typing.OnNext] = None, *,
                  scheduler: Optional[rx_typing.Scheduler] = None, param: Any = None) -> rx_typing.Disposable:

        # print(self.uid, observer, param)
        scheduler = scheduler or mape.rx_scheduler
        return super().subscribe(observer, on_error, on_completed, on_next, scheduler=scheduler)

    def _subscribe_core(self, observer, scheduler=None):
        """ Things to do on each subscribe """
        subscription = self._p_out.output.subscribe(observer, scheduler=scheduler)
        return CompositeDisposable(Disposable(lambda _: self.on_unsubscribe(observer)), subscription)

    def on_unsubscribe(self, observer):
        """ On each subscriber have disposed/ended the subscription """
        logger.debug(f"on_unsubscribe", observer)

    """ Start and stop element. 
    Only port_in stay readable, the rest is frozen (ie. no item transit).
    Good methods to extend for (de)allocate/start/stop internal element resources """

    def start(self, scheduler=None):
        if not self.is_running:
            # TODO: debug can be pre-pend here as ops.do_action() instead of subscribe ?!
            self._p_in.pipe = self._p_in.input.pipe(
                ops.filter(lambda item: not isinstance(item, typing.CallMethod)),
                ops.do_action(lambda item: isinstance(item, typing.Message) and item.add_hop(self)),
                *self._p_in.operators
            )

            sub_message = self._p_in.pipe.subscribe(
                lambda value: self._on_next(value, self._p_out.input.on_next),
                # self._on_next,
                self._on_error,
                self._on_completed,
                scheduler=scheduler
            )

            # TODO: implement the logic of do_action (ie. real call).
            # Maybe can be un external utility where pass self
            sub_method_call = self._p_in.input.pipe(
                ops.filter(lambda item: isinstance(item, typing.CallMethod)),
                ops.do_action(lambda item: item.exec(self))
            ).subscribe(scheduler=scheduler)

            self._p_in.disposable = CompositeDisposable(sub_message, sub_method_call)

            self._p_out.pipe = self._p_out.input.pipe(*self._p_out.operators)
            self._p_out.disposable = self._p_out.pipe.subscribe(self._p_out.output, scheduler=scheduler)

            self.is_running = True

        return Disposable(self.stop)

    def stop(self):
        if self.is_running:
            self._p_out.disposable.dispose()
            self._p_in.disposable.dispose()

            self.is_running = False

    def dispose(self) -> None:
        self.stop()
        # Uncomment (should) means that observer need to RE-subscribe
        # self._p_out.output.dispose()
        # self._p_in.input.dispose()
        # super().dispose()

    """ Called after [port in] => [pipe in] => [_on_next()] => [pipe out] => [port out] """

    def _on_next(self, value: Any, on_next: Optional[Callable] = None, *args, **kwargs) -> Any:
        """ Override to add your business logic """
        on_next = on_next or self._p_out.input.on_next
        on_next(value)

    # def _process_msg(self, value: Any move_on: Callable, *args, **kwargs):
    #     """ Subscribe for add your business logic """
    #     move_on(value)

    def __call__(self, value, *args, **kwargs):
        return self._on_next(value, on_next=self._p_out.input.on_next, *args, **kwargs)

    def _on_error(self, error: Exception) -> None:
        self._p_out.input.on_error(error)

    def _on_completed(self) -> None:
        self._p_out.input.on_completed()

    @property
    def uid(self) -> str:
        return self._uid

    def __str__(self) -> str:
        return self.uid

    @property
    def loop(self) -> mape.Loop:
        return self._loop

    @property
    def path(self) -> str:
        return f"{self.loop.uid}.{self.uid}"

    """ Port in and out aliases """

    @property
    def p_in(self):
        return self._p_in.input

    @property
    def p_out(self):
        return self._p_out.output

    @property
    def port_in(self):
        return self.p_in

    @property
    def port_out(self):
        return self.p_out

    @property
    def source(self):
        return self.p_in

    @property
    def sink(self):
        return self.p_out


class StartOnSubscribe(Element):
    def _subscribe_core(self, observer, scheduler=None):
        subscribe = super()._subscribe_core(observer, scheduler)
        super().start(scheduler=scheduler)

        return subscribe


class StartOnInit(Element):
    def __init__(self,
                 loop,
                 uid: str,
                 ops_in: Optional[Tuple] = (),
                 ops_out: Optional[Tuple] = (),
                 scheduler=None
                 ) -> None:
        super().__init__(loop, uid, ops_in, ops_out)
        super().start(scheduler=scheduler)


class Monitor(Element):
    prefix: str = 'm_'


class Analyze(StartOnSubscribe):
    prefix: str = 'a_'


class Plan(StartOnSubscribe):
    prefix: str = 'p_'


class Execute(StartOnInit):
    prefix: str = 'e_'


""" Allow display correct signature in IDE """
# ~/.local/share/JetBrains/IntelliJIdea2021.3/python/helpers/typeshed/stdlib/dataclasses.pyi

@overload
def to_element_cls(func: Callable) -> Type[Element]: ...


@overload
def to_element_cls(func: None) -> Callable[[Callable], Type[Element]]: ...


@overload
def to_element_cls(
        *,
        element_class=...,
        default_uid: str | UID = ...,
        default_ops_in: Optional[typing.OpsChain] = ...,
        default_ops_out: Optional[typing.OpsChain] = ...
) -> Type[Element]: ...


""" END  """


def to_element_cls(func=None, /, *,
                   element_class=Type[Element],
                   default_uid: str | UID = UID.DEF,
                   default_ops_in: Optional[typing.OpsChain] = (),
                   default_ops_out: Optional[typing.OpsChain] = ()
                   ) -> Type[Element] | Callable[..., Type[Element]]:
    """ Create the decorator and manage the call w/wo parentheses (ie @decorator vs @decorator()) """
    if func is None:
        # Called as @decorator(), with parentheses.
        # Return a function with all args set except 'func',
        # the second call (ie. @decorator(args)(func)) will set 'func'.
        return partial(to_element_cls,
                       element_class=element_class,
                       default_uid=default_uid,
                       default_ops_in=default_ops_in,
                       default_ops_out=default_ops_out)

    # Called as @decorator, without parentheses
    return make_func_class(func, element_class, default_uid, default_ops_in, default_ops_out)


# TODO: maybe can be passed *args, **kwargs (since default_uid)
def make_func_class(func: Callable | Coroutine,
                    element_class: Type[Element],
                    default_uid: str | UID = UID.DEF,
                    default_ops_in: Optional[typing.OpsChain] = (),
                    default_ops_out: Optional[typing.OpsChain] = ()
                    ) -> Type[Element]:

    if default_uid == UID.DEF:
        default_uid = func.__name__

    class ElementFunc(element_class):
        def __init__(self,
                     loop: mape.Loop,
                     uid: str | UID = default_uid,
                     ops_in: Optional[typing.OpsChain] = default_ops_in,
                     ops_out: Optional[typing.OpsChain] = default_ops_out
                     ) -> None:
            super().__init__(loop, uid, ops_in, ops_out)

            # Additional params/kwargs passed to func()
            self._on_next_opt_kwargs = {}

            # Discover what parameters (name and default) has function signature
            func_params = inspect.signature(func).parameters

            if 'self' in func_params.keys():
                self._on_next_opt_kwargs = {'self': self}

            # TODO: alternative _on_next definition to test
            # self.__class__._on_next = func

        @wraps(func)
        def _on_next(self, *args, **kwargs) -> Any | Awaitable:
            new_kwargs = {**self._on_next_opt_kwargs, **kwargs}

            # TODO: use mape.utils.auto_task
            if inspect.iscoroutinefunction(func):
                # The func execution is put in a task (ie parallel/background)
                coro = task_exception(func(*args, **new_kwargs))
                return self._aio_loop.create_task(coro)
            else:
                return func(*args, **new_kwargs)

        def add_param_to_on_next_call(self, kwargs):
            """ Pass a dict with key: value (as param=value) to pass during _on_next() calling. """
            self._on_next_opt_kwargs = kwargs

    # Code used when we act on object (and not class) level
    # Bound as a real "object method" passing self as first arg
    # func = types.MethodType(func, element)
    # setattr(element, '_on_next', func)
    # or: element._on_next = func

    return ElementFunc


# Dinamically add to module
# import sys
# sys.modules[__name__].ciao = "ciao"


to_monitor_cls = partial(to_element_cls, element_class=Monitor)
to_analyze_cls = partial(to_element_cls, element_class=Analyze)
to_plan_cls = partial(to_element_cls, element_class=Plan)
to_execute_cls = partial(to_element_cls, element_class=Execute)



from __future__ import annotations

import asyncio
import logging
import aioredis
from functools import partial

import rx
from rx.subject import Subject
from rx.core import Observer, Observable
from rx.disposable import Disposable
from rx import operators as ops

from typing import List

import mape
from mape.base_elements import Port
from mape.utils import log_task_exception
from .pubsub import subscribe_handler
from ..de_serializer import obj_to_raw, Pickled

logger = logging.getLogger(__name__)


class PubObserver(Observer):

    def __init__(self, channel, redis=None, serializer=None) -> None:
        self._channel = channel
        self._queue = asyncio.Queue()
        self._redis = redis or mape.redis
        self._serializer = serializer or partial(obj_to_raw, Pickled)

        if not isinstance(self._redis, aioredis.Redis):
            logger.error("You are trying to use Redis without config it!")

        self._p_in = Port(input=Subject(), operators=[ops.materialize()])
        self._p_in.pipe = self._p_in.input.pipe(*self._p_in.operators)
        self._p_in.disposable = self._p_in.pipe.subscribe(on_next=lambda item: self._publish(item))

        self._task = asyncio.create_task(self._publish_queue())

        super().__init__(self._p_in.input.on_next, self._p_in.input.on_error, self._p_in.input.on_completed)

    def _publish(self, item):
        self._queue.put_nowait(item)

    @log_task_exception
    async def _publish_queue(self):
        while True:
            item = await self._queue.get()
            await self._redis.publish(self._channel, self._serializer(item))

    def dispose(self) -> None:
        self._task.cancel()
        self._p_in.disposable.dispose()

    def __del__(self):
        self.dispose()


class SubObservable(Observable):
    def __init__(self, channels_patterns, deserializer=None, redis=None) -> None:
        self._channels_pattern = channels_patterns if isinstance(channels_patterns, List) else [channels_patterns]
        self._task = None

        def on_subscribe(observer, scheduler):
            sub_handlers = {pattern: observer.on_next for pattern in self._channels_pattern}
            self._task = subscribe_handler(sub_handlers, full_message=False, deserializer=deserializer, redis=redis)

            return Disposable(self.unsubscribe)

        self._auto_connect = rx.create(on_subscribe).pipe(ops.dematerialize(), ops.share())
        super().__init__()

    def _subscribe_core(self, observer, scheduler=None):
        return self._auto_connect.subscribe(observer, scheduler=scheduler)

    def unsubscribe(self):
        self._task.cancel()

    def __del__(self):
        self.unsubscribe()





from __future__ import annotations

import purse

from mape.remote.de_serializer import obj_from_raw, list_from_raw, obj_to_raw


def purse_monkey_patch():
    purse.collections._obj_from_raw = obj_from_raw
    purse.collections._list_from_raw = list_from_raw
    purse.collections._obj_to_raw = obj_to_raw



from __future__ import annotations

from datetime import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, List, Tuple, Callable, Optional, Union, Awaitable, Coroutine, NamedTuple, TypeVar

from rx.core.typing import Observer

from mape import base_elements
from mape.utils import aio_call

T1 = TypeVar('T1')
T2 = TypeVar('T2')

OpsChain = Union[Tuple, Callable]
Mapper = Callable[[T1], T2]
DestMapper = Callable[[T1], Union[List[Observer], Observer]]


@dataclass
class Item:
    src: str = None
    dst: str = None
    # TODO: or list?!
    hops: int = 0
    timestamp: float = field(default_factory=lambda: datetime.timestamp(datetime.now()))

    @staticmethod
    def _element2path(element: Optional[base_elements.Element | str]):
        return element.path if hasattr(element, 'path') else element

    def add_hop(self, hop):
        # save history of hops?!
        # hop.path()
        self.hops += 1


@dataclass
class Message(Item):
    value: Any = None

    @classmethod
    def create(cls, value, *,
               src: Optional[base_elements.Element | str] = None,
               dst: Optional[base_elements.Element | str] = None):
        return cls(value=value, src=cls._element2path(src), dst=cls._element2path(dst))

    def __repr__(self):
        formatted_time = datetime.fromtimestamp(self.timestamp).strftime('%H:%M:%S.%f')[:-3]
        return f"{self.__class__.__name__}({self.value}, {self.src}, {self.dst}, {self.hops}, {formatted_time})"


@dataclass
class CallMethod(Item):
    name: str = None
    args: tuple | list = ()
    kwargs: dict = field(default_factory=lambda: dict())

    @classmethod
    def create(cls, name, *args, **kwargs):
        return cls(name=name, args=args, kwargs=kwargs)

    def exec(self, obj_or_module):
        # TODO: test with async Method (maybe something like): ElementFunc._on_next()
        return getattr(obj_or_module, self.name)(*self.args, **self.kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.args}, {self.kwargs}, {self.src}, {self.dst}, {self.hops}, {self.timestamp})"


# TODO: delete... before check if somewhere is used
class MapeLoop(ABC):
    __slots__ = ()
    prefix: str = None

    @abstractmethod
    def __init__(self, uid: str = None, app=None) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_element(self, element):
        raise NotImplementedError

    @abstractmethod
    def has_element(self, element):
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, element):
        raise NotImplementedError

    @abstractmethod
    def __getattr__(self, uid):
        """ Allow access (through dot notation) to mape elements.
        note: __getattr__() is called only when no real object attr exist. """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, uid):
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

    @abstractmethod
    def uid(self):
        raise NotImplementedError

    @abstractmethod
    def elements(self):
        raise NotImplementedError

    @abstractmethod
    def monitor(self, func=None, /, *,
                uid=None, ops_in: Optional[OpsChain] = (), ops_out: Optional[OpsChain] = ()
                ) -> Any:
        raise NotImplementedError

    @abstractmethod
    def analyze(self, func=None, /, *,
                uid=None, ops_in: Optional[OpsChain] = (), ops_out: Optional[OpsChain] = ()
                ) -> Any:
        raise NotImplementedError

    @abstractmethod
    def plan(self, func=None, /, *,
             uid=None, ops_in: Optional[OpsChain] = (), ops_out: Optional[OpsChain] = ()
             ) -> Any:
        raise NotImplementedError

    @abstractmethod
    def execute(self,
                func=None,
                /, *,
                uid=None,
                ops_in: Optional[OpsChain] = (),
                ops_out: Optional[OpsChain] = ()
                ) -> Any:
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mape',
 'mape.remote',
 'mape.remote.influxdb',
 'mape.remote.redis',
 'mape.remote.rest']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Rx>=3.2.0,<4.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'fastapi>=0.73,<0.74',
 'influxdb-client>=1.26.0,<2.0.0',
 'redis-purse>=0.25.0,<0.26.0',
 'uvicorn>=0.17.4,<0.18.0']

setup_kwargs = {
    'name': 'pymape',
    'version': '0.1.0a3',
    'description': 'Framework to develop Self-Adaptive system based on MAPE-K loop.',
    'long_description': '<p align="center">\n    <a href="https://pypi.org/project/pymape/"><img\n        src="https://img.shields.io/pypi/v/pymape?style=flat-square"\n        alt="PyPI Version"\n    /></a>\n    <a href="https://pypi.org/project/pymape/"><img\n        src="https://img.shields.io/pypi/pyversions/pymape?style=flat-square"\n        alt="Py Version"\n    /></a>\n    <a href="https://github.com/elbowz/pymape/issues"><img\n        src="https://img.shields.io/github/issues/elbowz/pymape.svg?style=flat-square"\n        alt="Issues"\n    /></a>\n    <a href="https://raw.githubusercontent.com/elbowz/PyMAPE/main/LICENSE"><img\n        src="https://img.shields.io/github/license/elbowz/pymape.svg?style=flat-square"\n        alt="GPL License"\n    /></a>\n    <a href="https://raw.githubusercontent.com/elbowz/PyMAPE/main/LICENSE"><img\n        src="https://img.shields.io/static/v1?label=Powered&message=RxPY&style=flat-square&color=informational"\n        alt="RxPY"\n    /></a>\n</p>\n\n<p align="center">\n    <img src="https://github.com/elbowz/PyMAPE/raw/main/docs/img/logo.png" alt="PyMAPE" width="400">\n    <h4 align="center">Distributed and decentralized MonitorAnalyzePlanExecute-Knowledge loops framework</h3>\n    <p align="center">\n        Framework to support the development and deployment of Autonomous (Self-Adaptive) Systems\n    </p>\n</p>\n<br />\n\n## Getting Started\n\n### Install\n\n```bash\npip install pymape\n```\n\nSee [Examples](https://github.com/elbowz/PyMAPE#examples) for play with some MAPE-K patterns.\n\n### Install for Developers and Contributors\n\n```bash\ngit clone https://github.com/elbowz/PyMAPE.git\ncd PyMAPE\npoetry install\n```\n*note:* you need to have already installed [poetry](https://python-poetry.org/)\n\nThen use `poetry shell` and/or `poetry run` (eg. `poetry run examples/coordinated-ambulance.py --speed 80`) to exec your script inside the development environment.\n\n### First loop (Ambulance)\n\n![ambulance diagram](https://github.com/elbowz/PyMAPE/raw/main/docs/img/mape-ambulance.png)\n\n```python\nimport mape\nfrom mape.loop import Loop\n\n""" MAPE Loop and elements definition """\nloop = Loop(uid=\'ambulance_emergency\')\n\n@loop.monitor\ndef detect(item, on_next, self):\n    if \'speed_limit\' in item:\n        # Local volatile knowledge\n        self.loop.k.speed_limit = item[\'speed_limit\']\n    elif \'emergency_detect\' in item:\n        on_next(item[\'emergency_detect\'])\n\n@loop.plan(ops_in=ops.distinct_until_changed())\nasync def policy(emergency, on_next, self):\n    if emergency is True:\n        self.last_speed_limit = self.loop.k.speed_limit\n        new_speed = max(self.last_speed_limit, self.emergency_speed)\n\n        on_next({\'speed\': new_speed})\n        on_next({\'siren\': True})\n    else:\n        on_next({\'speed\': self.last_speed_limit})\n        on_next({\'siren\': False})\n\npolicy.emergency_speed = 160\n\n@loop.execute\ndef exec(item: dict, on_next):\n    if \'speed\' in item:\n        ambulance.speed_limit = item[\'speed\']\n    if \'siren\' in item:\n        ambulance.siren = item[\'siren\']\n\nfor element in loop:\n    element.debug(Element.Debug.IN)\n\n""" MAPE Elements connection """\ndetect.subscribe(policy)\npolicy.subscribe(exec)\n\n# Starting monitor...\ndetect.start()\n```\n### Traversing\n\n```python\n# Iterate over loops and element\nfor loop in mape.app:\n    logger.debug(f"* {loop.uid}")\n    for element in loop:\n        logger.debug(f" - {element.uid}")\n\n# Get all Execute elements\n[element for element in loop_obj if isinstance(element, Execute)]\n\n# Different access way to loop/element through dot-notation (path)\nmape.app.loop_uid.element_uid\nmape.app[\'loop_uid.element_uid\']\n```\n\n## Docs\n\n### Slides\n\n[Introduction to PyMAPE](https://github.com/elbowz/PyMAPE/raw/main/docs/slides.pdf) with examples\n\n### Examples\n\nImplementation of the 5 decentralized (and distributed) MAPE patterns described in the paper:  \n["On Patterns for Decentralized Control in Self-Adaptive Systems", Danny Weyns](https://www.ics.uci.edu/~seal/publications/2012aSefSAS.pdf)\n\n* **Ambulance-Car Emergency** (Information Sharing and Coordinated Control)\n* **Average Speed Enforcement** (Master/Slave)\n* **Dynamic Carriageway** (Regional Planning)\n* **Cruise Control with Distance Hold** (Hierarchical Control)\n\nIf you want try some examples (path `examples/`), refer to section `# CLI EXAMPLES` inside the source code of each one.  \n\nThe examples need furthers requirements, please see [pyproject.toml](https://github.com/elbowz/PyMAPE/raw/main/pyproject.toml) or use poetry to [install them](https://github.com/elbowz/PyMAPE#install-for-developers-and-contributors).  \n\nYou also need a Redis and InfluxDB instance running, for example:\n\n```bash\ndocker run --name mape-redis -p 6379:6379  \\\n-v $(pwd)/docker/redis:/usr/local/etc/redis  \\\n--rm redis redis-server /usr/local/etc/redis/redis.conf\n```\n\n```bash\ndocker run --name mape-influxdb -p 8086:8086 \\\n-v $(pwd)/docker/influxdb/data:/var/lib/influxdb2 \\\n-v $(pwd)/docker/influxdb/conf:/etc/influxdb2 \\\n-e DOCKER_INFLUXDB_INIT_MODE=setup \\\n-e DOCKER_INFLUXDB_INIT_USERNAME=user \\\n-e DOCKER_INFLUXDB_INIT_PASSWORD=qwerty123456 \\\n-e DOCKER_INFLUXDB_INIT_ORG=univaq \\\n-e DOCKER_INFLUXDB_INIT_BUCKET=mape \\\n-e DOCKER_INFLUXDB_INIT_RETENTION=1w \\\n-e DOCKER_INFLUXDB_INIT_ADMIN_TOKEN=<GENERATE_OR_TAKE_FROM_CONFIG_YAML> \\\n--rm influxdb:2.0\n```\n\nSee source for more information.',
    'author': 'Emanuele Palombo',
    'author_email': 'muttley.bd@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/elbowz/PyMAPE',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import asyncio

from uvicorn import Config, Server
from multiprocessing import Process

# TODO: format correctly the logger output


class UvicornDaemon:
    def __init__(self, app, loop, host='0.0.0.0', port=6060, log_level='info', **kwargs) -> None:
        self._loop = loop

        kwargs = {
            'app': app,
            'host': host,
            'port': int(port),
            'log_level': log_level,
            **kwargs
        }

        config = Config(**kwargs)
        self._server = Server(config)
        self._loop.create_task(self._server.serve())

    def stop(self):
        self._loop.create_task(self._server.shutdown())

    @property
    def config(self):
        return self._server.config

